# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class CachePolicy(Enum):
    NORMAL = 1,
    MINIMUM = 2,
    EAGER = 3,


_policy = CachePolicy.NORMAL


def set_cache_policy(policy):
    global _policy
    _policy = policy


MULTIPLIER = {
    CachePolicy.NORMAL: 1,
    CachePolicy.MINIMUM: 1 / 16,
    CachePolicy.EAGER: 2,
}


def get_entry_count(count):
    if count is None:
        if _policy == CachePolicy.MINIMUM:
            return 512
        return None

    multiplier = MULTIPLIER[_policy]
    return int(count * multiplier)
